﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Collections;

namespace Achiral.Extension
{
    public static partial class MakeWorld
    {
        public static IEnumerable<TResult> Select<T, TResult>(this T[,] source, Func<T, int, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");

            checked
            {
                for (int i0 = 0; i0 < source.GetLength(0); ++i0)
                {
                    for (int i1 = 0; i1 < source.GetLength(1); ++i1)
                    {
                        yield return selector(source[i0, i1], i0, i1);
                    }
                }
            }
        }
        public static IEnumerable<TResult> LongSelect<T, TResult>(this T[,] source, Func<T, long, long, TResult> selector)
        {
            AssertNonNullArgument(source, "source");

            checked
            {
                for (long i0 = 0; i0 < source.GetLongLength(0); ++i0)
                {
                    for (long i1 = 0; i1 < source.GetLongLength(1); ++i1)
                    {
                        yield return selector(source[i0, i1], i0, i1);
                    }
                }
            }
        }
        public static IEnumerable<TResult> Select<T, TResult>(this T[, ,] source, Func<T, int, int, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");

            checked
            {
                for (int i0 = 0; i0 < source.GetLength(0); ++i0)
                {
                    for (int i1 = 0; i1 < source.GetLength(1); ++i1)
                    {
                        for (int i2 = 0; i2 < source.GetLength(2); ++i2)
                        {
                            yield return selector(source[i0, i1, i2], i0, i1, i2);
                        }
                    }
                }
            }
        }
        public static IEnumerable<TResult> LongSelect<T, TResult>(this T[, ,] source, Func<T, long, long, long, TResult> selector)
        {
            AssertNonNullArgument(source, "source");

            checked
            {
                for (long i0 = 0; i0 < source.GetLongLength(0); ++i0)
                {
                    for (long i1 = 0; i1 < source.GetLongLength(1); ++i1)
                    {
                        for (long i2 = 0; i2 < source.GetLongLength(2); ++i2)
                        {
                            yield return selector(source[i0, i1, i2], i0, i1, i2);
                        }
                    }
                }
            }
        }
        
        public static void ForEach<T>(this IEnumerable<T> source, Action<T> action)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                action(item);
            }
        }
        public static void ForEach<T>(this IEnumerable<T> source, Action<T, int> action)
        {
            AssertNonNullArgument(source, "source");

            checked
            {
                int index = 0;
                foreach (var item in source)
                {
                    action(item, index);
                    ++index;
                }
            }
        }
        public static string Join(this IEnumerable<string> source, string separator)
        {
            AssertNonNullArgument(source, "source");

            var sb = new StringBuilder();
            source.ForEach((str, index) =>
            {
                if (index == 0)
                    sb.Append(str);
                else
                    sb.Append(separator).Append(str);
            });

            return sb.ToString();
        }
        public static IEnumerable<TResult> SelectWithFirst<T, TResult>(this IEnumerable<T> source, Func<T, T, TResult> func)
        {
            AssertNonNullArgument(source, "source");

            bool isFirst = true;
            var first = default(T);
            foreach (var item in source)
            {
                if (isFirst)
                {
                    first = item;
                    isFirst = false;
                }
                else
                {
                    yield return func(first, item);
                }
            }
        }

        public static IEnumerable<TResult> ZipWith<T1, T2, TResult>(this IEnumerable<T1> source1, IEnumerable<T2> source2, Func<T1, T2, TResult> func)
        {
            AssertNonNullArgument(source1, "source1");
            AssertNonNullArgument(source2, "source2");

            using (var enumerator1 = source1.GetEnumerator())
            using (var enumerator2 = source2.GetEnumerator())
            {
                while (enumerator1.MoveNext() && enumerator2.MoveNext())
                {
                    yield return func(enumerator1.Current, enumerator2.Current);
                }
            }
        }

        public static IEnumerable<TResult> ZipWith<T1, T2, TResult>(this IEnumerable<T1> source1, IEnumerable<T2> source2, Func<T1, T2, int, TResult> func)
        {
            AssertNonNullArgument(source1, "source1");
            AssertNonNullArgument(source2, "source2");

            using (var enumerator1 = source1.GetEnumerator())
            using (var enumerator2 = source2.GetEnumerator())
            {
                int index = 0;
                while (enumerator1.MoveNext() && enumerator2.MoveNext())
                {
                    yield return func(enumerator1.Current, enumerator2.Current, index);
                    index = checked(index + 1);
                }
            }            
        }

        /// <see>http://d.hatena.ne.jp/siokoshou/20071227#c1198746342</see>
        public static IEnumerable<T> CascadeDepthFirst<T>(this IEnumerable<T> source, Func<T, IEnumerable<T>> func)
        {
            AssertNonNullArgument(source, "source");

            var enumeratorStack = new Stack<IEnumerator<T>>();
            try
            {
                var top = source.GetEnumerator();
                while (true)
                {
                    if (top.MoveNext())
                    {
                        yield return top.Current;
                        enumeratorStack.Push(top);
                        top = func(top.Current).GetEnumerator();
                    }
                    else
                    {
                        if (enumeratorStack.Count <= 0)
                        {
                            yield break;
                        }
                        top = enumeratorStack.Pop();
                    }
                }
            }
            finally
            {
                enumeratorStack.ForEach(e => e.Dispose());
            }
        }
        public static IEnumerable<T> CascadeDepthFirst<T>(this IEnumerable<T> source, ILookup<T, T> lookup)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(lookup, "lookup");

            return source.CascadeDepthFirst(lookup.AsFunc());
        }
        public static IEnumerable<TResult> CascadeDepthFirst<T, TResult>(this IEnumerable<T> source, Func<T, IEnumerable<T>> func, Func<T, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");

            var enumeratorStack = new Stack<IEnumerator<T>>();
            try
            {
                var top = source.GetEnumerator();
                while (true)
                {
                    if (top.MoveNext())
                    {
                        yield return selector(top.Current, enumeratorStack.Count);
                        enumeratorStack.Push(top);
                        top = func(top.Current).GetEnumerator();
                    }
                    else
                    {
                        if (enumeratorStack.Count <= 0)
                        {
                            yield break;
                        }
                        top = enumeratorStack.Pop();
                    }
                }

            }
            finally
            {
                enumeratorStack.ForEach(e => e.Dispose());
            }
        }
        public static IEnumerable<TResult> CascadeDepthFirst<T, TResult>(this IEnumerable<T> source, ILookup<T, T> lookup, Func<T, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(lookup, "lookup");

            return source.CascadeDepthFirst(lookup.AsFunc(), selector);
        }
        public static IEnumerable<T> CascadeBreadthFirst<T>(this IEnumerable<T> source, Func<T, IEnumerable<T>> func)
        {
            AssertNonNullArgument(source, "source");

            var list = source.ToList();
            int nestLevel = 0;
            while (true)
            {
                if (!list.Any()) yield break;
                foreach (var item in list) yield return item;
                ++nestLevel;
                list = list.SelectMany(item => func(item)).ToList();
            }
        }
        public static IEnumerable<T> CascadeBreadthFirst<T>(this IEnumerable<T> source, ILookup<T, T> lookup)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(lookup, "lookup");

            return source.CascadeBreadthFirst(lookup.AsFunc());
        }

        public static IEnumerable<TResult> CascadeBreadthFirst<T, TResult>(this IEnumerable<T> source, Func<T, IEnumerable<T>> func, Func<T, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");

            var list = source.ToList();
            int nestLevel = 0;
            while (true)
            {
                if (!list.Any()) yield break;
                foreach (var item in list) yield return selector(item, nestLevel);
                ++nestLevel;
                list = list.SelectMany(item => func(item)).ToList();
            }
        }
        public static IEnumerable<TResult> CascadeBreadthFirst<T, TResult>(this IEnumerable<T> source, ILookup<T, T> lookup, Func<T, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(lookup, "lookup");

            return source.CascadeBreadthFirst(lookup.AsFunc(), selector);
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(item);
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, string category)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(item, category);
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, string> selector)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(selector(item));
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, string> selector, string category)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(selector(item), category);
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, int, string> selector)
        {
            AssertNonNullArgument(source, "source");

            checked
            {
                int index = 0;
                foreach (var item in source)
                {
                    System.Diagnostics.Trace.WriteLine(selector(item, index));
                    yield return item;
                    ++index;
                }
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, int, string> selector, string category)
        {
            AssertNonNullArgument(source, "source");

            checked
            {
                int index = 0;
                foreach (var item in source)
                {
                    System.Diagnostics.Trace.WriteLine(selector(item, index), category);
                    yield return item;
                    ++index;
                }
            }
        }

        public static void ConsoleWriteLine<T>(this IEnumerable<T> source)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                Console.WriteLine(item);
            }
        }
        public static void ConsoleWriteLine<T>(this IEnumerable<T> source, Func<T, string> messageSelector)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                Console.WriteLine(messageSelector(item));
            }
        }
        public static void ConsoleWriteLine<T>(this IEnumerable<T> source, string format, Func<T, object[]> argumentsSelector)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                Console.WriteLine(format, argumentsSelector(item));
            }
        }
        public static void ConsoleWriteLine<T>(this IEnumerable<T> source, string format, params Func<T, object>[] selectors)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                Console.WriteLine(format, selectors.Select(func => func(item)));
            }
        }

        public static IEnumerable<T> Hook<T>(this IEnumerable<T> source, EnumeratorHook<T> hook)
        {
            AssertNonNullArgument(source, "source");

            return hook.Hook(source);
        }
    }

    public static partial class MakeWorld
    {
        public static ILookup<TKey, TValue> ToLookup<TSource, TKey, TValue>(this IEnumerable<TSource> source, Func<TSource, Func<TKey, bool>> patternSelector, Func<TSource, TValue> valueSelector)
        {
            AssertNonNullArgument(source, "source");
            return new PatternMatchedLookup<TKey, TValue>
            (
                source.Select(item => new MapElement<TKey, TValue>(patternSelector(item), valueSelector(item)))
                      .ToArray()
            );
        }
        private partial struct MapElement<TKey, TValue>
        {
            public readonly Func<TKey, bool> Pattern;
            public readonly TValue Value;
            public MapElement(Func<TKey, bool> pattern, TValue value)
            {
                this.Pattern = pattern;
                this.Value = value;
            }
        }
        private sealed partial class PatternMatchedLookup<TKey, TValue> : ILookup<TKey, TValue>
        {
            private readonly IList<MapElement<TKey, TValue>> _patterns;
            public PatternMatchedLookup(IList<MapElement<TKey, TValue>> patterns)
            {
                _patterns = patterns;
            }

            public bool Contains(TKey key)
            {
                return _patterns.Any(pattern => pattern.Pattern(key));
            }

            public int Count
            {
                get { return _patterns.Count; }
            }

            public IEnumerable<TValue> this[TKey key]
            {
                get
                {
                    return _patterns.Where(pattern => pattern.Pattern(key))
                                    .Select(pattern => pattern.Value);
                }
            }
            public IEnumerator<IGrouping<TKey, TValue>> GetEnumerator()
            {
                throw new NotImplementedException();
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }
        }
    }

    public struct EnumeratorHook<T>
    {
        public Action OnBeforeBegin { set; internal get; }
        public Action<IEnumerator<T>> OnAfterBegin { set; internal get; }
        public Action<IEnumerator<T>> OnBeforeReset { set; internal get; }
        public Action<IEnumerator<T>> OnAfterReset { set; internal get; }
        public Action<IEnumerator<T>> OnBeforeGetCurrent { set; internal get; }
        public Action<IEnumerator<T>, T> OnAfterGetCurrent { set; internal get; }
        public Action<IEnumerator<T>> OnBeforeMoveNext { set; internal get; }
        public Action<IEnumerator<T>, bool> OnAfterMoveNext { set; internal get; }
        public Action<IEnumerator<T>> OnBeforeDispose { set; internal get; }
        public Action<IEnumerator<T>> OnAfterDispose { set; internal get; }

        public IEnumerable<T> Hook(IEnumerable<T> source)
        {
            MakeWorld.AssertNonNullArgument(source, "source");

            return new MakeWorld.HookableEnumerable<T>(source, this);
        }
    }

    public static partial class MakeWorld
    {
        internal class HookableEnumerable<T> : IEnumerable<T>
        {
            private readonly IEnumerable<T> _source;
            private readonly EnumeratorHook<T> _hook;

            public HookableEnumerable(IEnumerable<T> source, EnumeratorHook<T> hook)
            {
                this._source = source;
                this._hook = hook;
            }
            public IEnumerator<T> GetEnumerator()
            {
                return new HookableEnumerator<T>(_source, _hook);
            }
            System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }
        }
        private class HookableEnumerator<T> : IEnumerator<T>
        {
            private readonly EnumeratorHook<T> _hook;
            private readonly IEnumerator<T> _source;

            public HookableEnumerator(IEnumerable<T> source, EnumeratorHook<T> hook)
            {
                _hook = hook;
                _hook.OnBeforeBegin.InvokeOrNop();
                _source = source.GetEnumerator();
                _hook.OnAfterBegin.InvokeOrNop(_source);
            }

            public T Current
            {
                get
                {
                    _hook.OnBeforeGetCurrent.InvokeOrNop(_source);
                    var current = _source.Current;
                    _hook.OnAfterGetCurrent.InvokeOrNop(_source, current);
                    return current;
                }
            }
            public void Dispose()
            {
                _hook.OnBeforeDispose.InvokeOrNop(_source);
                _source.Dispose();
                _hook.OnAfterDispose.InvokeOrNop(_source);
            }
            object System.Collections.IEnumerator.Current
            {
                get { return this.Current; }
            }
            public bool MoveNext()
            {
                _hook.OnBeforeMoveNext.InvokeOrNop(_source);
                bool result = _source.MoveNext();
                _hook.OnAfterMoveNext.InvokeOrNop(_source, result);
                return result;
            }
            public void Reset()
            {
                _hook.OnBeforeReset.InvokeOrNop(_source);
                _source.Reset();
                _hook.OnAfterReset.InvokeOrNop(_source);
            }
        }
    }
}